/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-07 * implemented rather crappy bitmask compression - Rolf
03/Mar-06 * started using new BitStream utilites for compression - Rolf
03/Mar-06 * switched to huffman encoding for image data - Rolf
03/Mar-05 * kicked old save_to_file, reimplemented save, added FILE* constructor - Rolf
03/Mar-05 * added proper LGPL notice - Rolf
03/Mar-03 * reintroduced 'stride' for the bit mask
          * added move function (preliminary, see below)
          * kicked the fancy 'delta compression' and went back to plain RLE - Rolf

*/


#include <vector>			//for MaskedImage.h
#include "common.h"
#include "Span.h"			//for MaskedImage.h
#include "Section.h"		//for MaskedImage.h

#include "MaskedImage.h"

//internal stuff
#include <stdlib.h>
#include <memory.h>
#include <stdio.h>
#include "Counter2D.h"
#include "x86_timer.ah"
#include "Timing.h"
#include "Huffman.h"
#include "BitStream.h"

HuffmanCode huff;


MaskedImage::MaskedImage(int width,int height):
width(width),height(height),
compressed_image(NULL),compressed_mask(NULL),
compressed_size(0),compressed_mask_size(0)
{
    stride=(width+63)>>3;
	mask=(ubyte*)calloc(height*stride,sizeof(ubyte));
	image=(ubyte*)calloc(width*height,sizeof(ubyte));
}

MaskedImage::MaskedImage(FILE* inf):
image(NULL),compressed_image(NULL),
mask(NULL),compressed_mask(NULL),
width(0),height(0),stride(0),
compressed_size(0),compressed_mask_size(0)
{
	char version_string[16];
	uint version;
	fgets(version_string,16,inf);
	if (('v'!=version_string[0])||('\n'!=version_string[4]))
	{
		printf("Puke!\n");
	//FIXME
	}
	else
	{
		version=atoi(version_string+1);

		fread(&width,sizeof(int),1,inf);
		fread(&height,sizeof(int),1,inf);
		fread(&compressed_size,sizeof(int),1,inf);

		stride=(width+63)>>3;
		compressed_image=(ubyte*)malloc(compressed_size);

		if (version>=1)		//version 001 is the first to use mask compression
		{
			mask=NULL;
			fread(&compressed_mask_size,sizeof(int),1,inf);
			compressed_mask=(ubyte*)malloc(compressed_mask_size);
			fread(compressed_mask,1,compressed_mask_size,inf);
   		}
   		else				//version 000 only compressed the image itself
   		{
   			compressed_mask_size=0;
   			compressed_mask=NULL;
   			mask=(ubyte*)malloc(height*stride);
   			fread(mask,stride,height,inf);
        }
		fread(compressed_image,compressed_size,sizeof(ubyte),inf);
//convert old style RLE compression to new format
		if (0==version)
		{
			decompress_v000();
			compress();
		}
	}
}

MaskedImage::~MaskedImage()
{
	SAFE_FREE(image);
	SAFE_FREE(mask);
	SAFE_FREE(compressed_image);
	
}

bool
MaskedImage::save(FILE* of)
{
	compress();		//try to save disk space
	fputs("v001\n",of);
	fwrite(&width,sizeof(int),1,of);
	fwrite(&height,sizeof(int),1,of);
	fwrite(&compressed_size,sizeof(int),1,of);
	fwrite(&compressed_mask_size,sizeof(int),1,of);
	fwrite(compressed_mask,1,compressed_mask_size,of);
	fwrite(compressed_image,compressed_size,sizeof(ubyte),of);

	return(true);
}

void
MaskedImage::clear_mask()const
{
	memset(mask,0,height*stride*sizeof(ubyte));
}

MaskedImage::MaskedImage(const MaskedImage& rhs):
width(rhs.width),
height(rhs.height),
stride(rhs.stride),
compressed_size(rhs.compressed_size),
image(NULL),compressed_image(NULL),
mask(NULL),compressed_mask(NULL)
{
//delegate to assignment
	*this=rhs;
}

MaskedImage&
MaskedImage::operator =(const MaskedImage& rhs)
{
	if (this==&rhs) return(*this);

	width=rhs.width;	height=rhs.height;
	stride=rhs.stride;
	SAFE_FREE(mask);
	SAFE_FREE(image);
	SAFE_FREE(compressed_image);
	SAFE_FREE(compressed_mask);
	
	compressed_size=rhs.compressed_size;
	compressed_mask_size=rhs.compressed_mask_size;

	if (rhs.image)
	{
		image=(ubyte*)malloc(width*height*sizeof(ubyte));
 		memcpy(image,rhs.image,width*height*sizeof(ubyte));
	}
	if (rhs.mask)
  	{
		mask=(ubyte*)malloc(stride*height*sizeof(ubyte));
		memcpy(mask,rhs.mask,stride*height*sizeof(ubyte));
    }
	if (rhs.compressed_image)
	{
		compressed_image=(ubyte*)malloc(compressed_size);
 		memcpy(compressed_image,rhs.compressed_image,compressed_size);
   	}
	if (rhs.compressed_mask)
	{
		compressed_mask=(ubyte*)malloc(compressed_mask_size);
 		memcpy(compressed_mask,rhs.compressed_mask,compressed_mask_size);
	}

	return(*this);
}

void
MaskedImage::fill(const Span& span,const int value) const
{
	ubyte* i=image+width*span.y;
	for (int x=span.x0;x<=span.x1;++x) i[x]=value;
}

void
MaskedImage::fill(const std::vector <Span>& spans,const int value) const
{
	for (std::vector<Span>::const_iterator i=spans.begin();i!=spans.end();++i)
	{
		memset(image+width*i->y+i->x0,value,i->x1-i->x0);
	}
}

void
MaskedImage::fill(const Section& rect,const uint value) const
{
	for (uint y=rect.y0;y<=rect.y1;++y)
	{
		for (uint x=rect.x0;x<=rect.x1;++x)
		{
			image[width*y+x]=value;
		}
	}
}

uint
MaskedImage::fill_interior(const Section& rect,const uint value) const
{
	uint ret=0;
	int len=rect.x1-rect.x0-1;
	ubyte* i=image+width*(rect.y0+1)+rect.x0+1;
	for (int y=rect.y0+1;y<rect.y1;++y)
	{
		memset(i,value,len);
		i+=width;
		ret+=len;
	}
	return(ret);
}

void
MaskedImage::preview_outline(const Section& rect,const uint value) const
{
	for (int x=rect.x0;x<=rect.x1;++x)
	{
		if (!check_mask(x,rect.y0)) image[width*rect.y0+x]=value;
		if (!check_mask(x,rect.y1)) image[width*rect.y1+x]=value;
	}
	for (int y=rect.y0+1;y<rect.y1;++y)
	{
		if (!check_mask(rect.x0,y)) image[width*y+rect.x0]=value;
		if (!check_mask(rect.x1,y)) image[width*y+rect.x1]=value;
	}
}

void
MaskedImage::preview_fill(const Section& rect,const uint value) const
{
	for (int y=rect.y0;y<=rect.y1;++y)
	{
		for (int x=rect.x0;x<=rect.x1;++x)
		{
			if (!check_mask(x,y)) image[width*y+x]=value;
		}
	}
}

void
MaskedImage::compress()
{
	if (NULL==image) return;
	SAFE_FREE(compressed_image);

//	AbsTimer timer;
//	printf("compressing from %u to ... ",get_used_memory_size());

	ubyte last;				//previous sample (we compress deltas)
	ubyte c=0;

	Counter2D i(width,height);
	BitSink bits(compressed_image);

	while ((!i.through())&&(0==check_mask(i.x,i.y))) ++i;

	while (!i.through())
	{
		last=c;
		c=image[i.linear()];
		
		const TreeSymbol* t=huff.get_code(c-last);
		bits.insert(t->code,t->code_len);

		do 
		{
			++i;
		} while ((!i.through())&&(0==check_mask(i.x,i.y)));
	}
	compressed_size=bits.finalize();
	
	compress_mask();

	SAFE_FREE(image);
//	printf("%u bytes - %.3f seconds\n",get_used_memory_size(),timer.elapsed_seconds());
}

void
MaskedImage::decompress()
{
	if (image) return;
	SAFE_FREE(image);
	image=(ubyte*)calloc(width*height,1);
//	AbsTimer timer;

	decompress_mask();
	Counter2D o(width,height);
	BitStream bits(compressed_image,compressed_size);

	while ((!o.through())&&(0==check_mask(o.x,o.y))) ++o;

	ubyte c=0;
	while ((!o.through())&&(!bits.over()))
	{
		const TreeSymbol* t=huff.extract_symbol(bits.peek());

		bits.advance(t->code_len);
		c+=t->value;
		do
		{
			image[o.linear()]=c;
			++o;
		} while ((!o.through())&&(0==check_mask(o.x,o.y)));
	}
	SAFE_FREE(compressed_image);
	compressed_size=0;

//	printf("decompressed - %.3f seconds\n",timer.elapsed_seconds());
}

void
MaskedImage::decompress_v000()
{
	if (image) return;
	SAFE_FREE(image);
	image=(ubyte*)calloc(width*height,1);

	Counter2D o(width,height);
	int i=0;
	uint len;

	while ((!o.through())&&(0==check_mask(o.x,o.y))) ++o;

	while ((!o.through())&&(i<compressed_size))
	{
		ubyte c=compressed_image[i];
		++i;
		if ((c&0xE0)==0x40)
		{
			len=(c&31)+1;
			if (i==compressed_size) break;
			c=compressed_image[i];
			++i;
		}
		else
			len=1;

		for (uint run=0;run<len;++run)
		{
			do
			{
				image[o.linear()]=c;
				++o;
			} while ((!o.through())&&(0==check_mask(o.x,o.y)));
		}
	}
	SAFE_FREE(compressed_image);
	compressed_size=0;
}	

const ubyte*
MaskedImage::get_data()
{
	decompress();
	return(image);
}

int
MaskedImage::get_used_memory_size()const
{
	int ret=sizeof(MaskedImage);
	if (image) ret+=width*height*sizeof(ubyte);
	if (compressed_image) ret+=compressed_size;
	ret+=height*stride*sizeof(ubyte);
	return(ret);
}

int
MaskedImage::estimate_required_memory(int width,int height)
{
	int ret=sizeof(MaskedImage);
	ret+=sizeof(ubyte)*width*height;			//uncompressed image
	ret+=height*((width+63)>>3)*sizeof(ubyte);   //mask
	return(ret);
}

void
MaskedImage::move(int delta_x,int delta_y)
{
    if ((delta_x==0)&&(delta_y==0)) return;
    decompress();
    int min_y=0;
    int max_y=height-1;
    if (delta_y!=0)
    {
        if (delta_y>0)
        {
            memmove(mask,mask+stride*delta_y,stride*(height-delta_y)*sizeof(ubyte));
            memset(mask+stride*(height-delta_y),0,stride*delta_y*sizeof(ubyte));
            memmove(image,image+width*delta_y,width*(height-delta_y)*sizeof(ubyte));
            memset(image+width*(height-delta_y),0,width*delta_y*sizeof(ubyte));
            max_y-=delta_y;
        }
        else        //delta_y<0
        {
            memmove(mask-stride*delta_y,mask,stride*(height+delta_y)*sizeof(ubyte));
            memset(mask,0,stride*(-delta_y)*sizeof(ubyte));
            memmove(image-width*delta_y,image,width*(height+delta_y)*sizeof(ubyte));
            memset(image,0,width*(-delta_y)*sizeof(ubyte));
            min_y-=delta_y;
        }
    }
    if (delta_x!=0)
    {
        bool no_shift=(0==(delta_x&7));
        if (delta_x>0)
        {
            for (int y=min_y;y<=max_y;++y)
            {
                memmove(image+width*y,image+width*y+delta_x,width-delta_x);
                memset(image+width*(y+1)-delta_x,0,delta_x);
//HACK: assumes no_shift==true
				memmove(mask+stride*y,mask+stride*y+(delta_x>>3),stride-(delta_x>>3));
				memset(mask+stride*y+((width-delta_x)>>3),0,delta_x>>3);
            }
        }
        else    //delta_x<0
        {
            for (int y=min_y;y<=max_y;++y)
            {
                memmove(image+width*y-delta_x,image+width*y,width+delta_x);
                memset(image+width*y,0,-delta_x);
//HACK: assumes no_shift==true
				memmove(mask+stride*y-(delta_x>>3),mask+stride*y,(width+delta_x)>>3);
				memset(mask+stride*y,0,(-delta_x)>>3);
            }
        }
    }
}

void
MaskedImage::compress_mask()
{
	uint out=0;
	
	SAFE_FREE(compressed_mask);
	BitSink bits(compressed_mask);

	Counter2D i(stride,height);

	uint last=0;

	while (!i.through())
	{
		uint run=0;
		uint x=mask[i.linear()];
		++i;
		if (x==last)
		{
			bits.insert(0,1);
		}
		else
			bits.insert(1+(x<<1),9);

		last=x;
	}
	compressed_mask_size=bits.finalize();
	SAFE_FREE(mask);
}

void
MaskedImage::decompress_mask()
{
	SAFE_FREE(mask);
	mask=(ubyte*)calloc(height*stride,1);

	uint last=0;

	Counter2D o(stride,height);
	BitStream bits(compressed_mask,compressed_mask_size);
	while ((!o.through())&&(!bits.over()))
	{
		uint x=bits.peek();
		if ((x&1)==1)
		{
			last=(x>>1)&255;
			bits.advance(9);
		}
		else
			bits.advance(1);

		mask[o.linear()]=last;
		++o;
	}
	SAFE_FREE(compressed_mask);
}
