/*
this is part of brot2, an example client to the brot_core library
Copyright (C) 2003 Rolf Neuberger

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added save_all, save_minimum and load - Rolf
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/


//class wrapper for the brot core DLL interface

#ifndef BC_CONTOUR_TRACING
#error You must first include brot_core_interface.h
#endif


class
BrotCoreContext
{
private:
	BrotCoreContext(); //hide default constructor
public:
	BrotCoreContext(int megs) {bcc=bc_create_context(megs);}
	~BrotCoreContext() {bc_delete_context(&bcc);}
	//reset the whole context
	void reset_all() {bc_reset_context(bcc);}

	bool load(const char* filename) {return(bc_load(bcc,filename));}
	bool save_all(const char* filename) {return(bc_save_all(bcc,filename));}
	bool save_minimum(const char* filename) {return(bc_save_minimum(bcc,filename));}

////misc
	//return string description of a generator type, given by enum
	static const char* const describe_generator(int type) {return(bc_describe_generator(type));}

////control output image
	//adjust image dimensions, aa strength and gamma for downfiltering
	//pass negative values for stuff you don't want to change
	void set_image_properties(int width,int height,int aa_strength,float aa_gamma) {
		bc_set_image_properties(bcc,width,height,aa_strength,aa_gamma);}
	//set a new 256 color rgb palette (768 bytes). The palette is copied, so it may be deleted from client memory
	void set_palette(const ubyte* palette) { bc_set_palette(bcc,palette); }
	//set palette cycling speed in palette entries per second. Both positive and negative values are valid
	void set_palette_cycling_speed(float speed) {bc_set_palette_cycling_speed(bcc,speed);}
	//set a new generator (=optimization strategy) by enumerant
	int set_generator_type(int type) { return(bc_set_generator_type(bcc,type)); }
	//change generator type, so the client does not need to know how many there are, and which one is current
	int change_generator_type() { return(bc_change_generator_type(bcc)); }
	//toggle preview; return false if disabled/true if enabled
	bool toggle_preview() { return(bc_toggle_preview(bcc)); }

////crunch and display
	//refine current view. returns false if the image is finished.
	bool refine_image() { return(bc_refine_image(bcc)); }
	//start over with the current view
	void reset_current_view() { bc_reset_current_view(bcc); }
	//get full color output image, already downfiltered if aa is active
	const ubyte* get_rgb_image() { return(bc_get_rgb_image(bcc)); }
	//get width/height of _downfiltered_ image
	int get_width() { return(bc_get_image_width(bcc)); }
	int get_height() { return(bc_get_image_height(bcc)); }
	//get amount of samples that could be saved due to optimizations; range [0...1]
	float get_cleverness() { return(bc_get_cleverness(bcc)); }
	//get _estimated_ progress for the current view; range [0...1]
	float get_progress() { return(bc_get_progress(bcc)); }
	//get time (in seconds) spent crunching on the current view
	float get_time_spent() { return(bc_get_time_spent(bcc)); }
	//refine other images contained in the history. no statistics available for these ...
	bool refine_history() { return(bc_refine_history(bcc)); }

////control underlying fractal
	//set new parameters for the mandelbrot fractal. only regular doubles => not _terribly_ precise
	void set_view_params(double real,double img,double radius) {
		bc_set_view_params_from_doubles(bcc,real,img,radius); }
	//set new params from c-strings. parses ExtremeFloats, integers and regular doubles output with sprintf
	//for the radius, regular floating point is always enough
	void set_view_params(char* real,char* img,double radius) {
		bc_set_view_params_from_text(bcc,real,img,radius); }
	//set params from a 'Place' struct. full precision
	void set_view_params(Place* p) { bc_restore_location_from(bcc,p); }
	//apply a scaling factor to the current max iteration depth, and return the updated value
	uint adjust_maximum_iterations(float factor) { return(bc_adjust_maximum_iterations(bcc,factor)); }

////get fractal parameter info
	//grab center real/imaginary and radius in double precision floating point format
	void get_view_params(double* real,double* img,double* radius) {
		bc_get_view_params_as_doubles(bcc,real,img,radius); }
	//ditto, but write text strings. this one exposes the full internal precision.
	//buffers sizes for real/img should be at least 64 chars each
	void get_view_params(char* real,char* img,double* radius) {
		bc_get_view_params_as_text(bcc,real,img,radius); }
	//store to a 'Place'. That's a little struct, preserves full precision
	void get_view_params(Place* p) { bc_store_location_to(bcc,p); }
	//get the current maximum iteration depth
	uint get_maximum_iterations() { return(bc_get_maximum_iterations(bcc)); }

////navigation
	//zoom on image coords. 0/0 is top left. factors >1 zoom in, factors<1 zoom out
	void zoom(int x,int y,float factor) { bc_zoom_on_window_coord(bcc,x,y,factor); }
	//ditto, but always zoom on image center
	void zoom(float factor) { bc_zoom_centered(bcc,factor); }
	//move the current view around. deltas are in units of screen size
	//eg delta_x=+0.5 moves half an image width to the right
	void move(float delta_x,float delta_y) { bc_move(bcc,delta_x,delta_y); }

//history
	//go back to last view, return false if we couldn't go back
	bool history_back() { return(bc_history_back(bcc)); }
	//go forward again (prolly after first going back)
	bool history_forward() { return(bc_history_forward(bcc)); }




private:
	bc_context bcc;		//identifies this context with the dll
};

