/*
this is the interface to brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

#define BC_CONTOUR_TRACING 0
#define BC_BOX_MATCHING 1
#define BC_BRUTE_FORCE 2
#define BC_MAX_GENERATOR_TYPE BC_BRUTE_FORCE

#ifdef BROT_CORE_EXPORTS
#define BROT_CORE_API __declspec(dllexport)
#else
#define BROT_CORE_API __declspec(dllimport)
#endif

#define BROT_CALL 

#ifdef BROT_CORE_EXPORTS
class Context;
typedef Context* bc_context;
#else
typedef unsigned int bc_context;
struct Place {ubyte dont_touch[100];};
#endif


#define BROT_FUNCTION(name,return_type,argument_list) BROT_CORE_API return_type BROT_CALL name argument_list

extern "C"
{
////context management
//create a new context for the core functions
//will try to consume no more than 'megs' megabytes of memory for the history functions
BROT_FUNCTION(bc_create_context,bc_context,(int megs));
//reset context to fresh state, but keep it around
BROT_FUNCTION(bc_reset_context,void,(bc_context context));
//destroy context
BROT_FUNCTION(bc_delete_context,void,(bc_context* context));

////misc
//return string description of a generator type, given by enum
BROT_FUNCTION(bc_describe_generator,const char* const,(int generator_type));

////control output image
//adjust image dimensions, aa strength and gamma for downfiltering
//pass negative values for stuff you don't want to change
BROT_FUNCTION(bc_set_image_properties,void,(bc_context c,int width,int height,int aa_level,float aa_gamma));
//set a new 256 color rgb palette (768 bytes). The palette is copied, so it may be deleted from client memory
BROT_FUNCTION(bc_set_palette,void,(bc_context c,const ubyte* const palette));
//set palette cycling speed in palette entries per second. Both positive and negative values are valid
BROT_FUNCTION(bc_set_palette_cycling_speed,void,(bc_context c,float speed));
//set a new generator (=optimization strategy) by enumerant
BROT_FUNCTION(bc_set_generator_type,int,(bc_context c,int type));
//change generator type, so the client does not need to know how many there are, and which one is current
BROT_FUNCTION(bc_change_generator_type,int,(bc_context c));
//toggle preview; return false if disabled/true if enabled
BROT_FUNCTION(bc_toggle_preview,bool,(bc_context c));

////crunch and display
//refine current view. returns false if the image is finished.
BROT_FUNCTION(bc_refine_image,bool,(bc_context c));
//start over
BROT_FUNCTION(bc_reset_current_view,void,(bc_context c));
//get full color output image, already downfiltered if aa is active
BROT_FUNCTION(bc_get_rgb_image,const ubyte*,(bc_context c));
//get width/height of _downfiltered_ image
BROT_FUNCTION(bc_get_image_width,int,(bc_context c));
BROT_FUNCTION(bc_get_image_height,int,(bc_context c));
//get amount of samples that could be saved due to optimizations; range [0...1]
BROT_FUNCTION(bc_get_cleverness,float,(bc_context c));
//get _estimated_ progress for the current view; range [0...1]
BROT_FUNCTION(bc_get_progress,float,(bc_context c));
//get time (in seconds) spent crunching on the current view
BROT_FUNCTION(bc_get_time_spent,float,(bc_context c));
//refine other images contained in the history. no statistics available for these ...
BROT_FUNCTION(bc_refine_history,bool,(bc_context c));


////control underlying fractal
//set new parameters for the mandelbrot fractal. only regular doubles => not _terribly_ precise
BROT_FUNCTION(bc_set_view_params_from_doubles,void,(bc_context c,double real,double img,double radius));
//set new params from c-strings. parses ExtremeFloats, integers and regular doubles output with sprintf
//for the radius, regular floating point is always enough
BROT_FUNCTION(bc_set_view_params_from_text,void,(bc_context c,char* real,char* img,double radius));
//get params from a 'Place' struct. full precision
BROT_FUNCTION(bc_restore_location_from,void,(bc_context c,void* place));
//apply a scaling factor to the current max iteration depth, and return the updated value
BROT_FUNCTION(bc_adjust_maximum_iterations,uint,(bc_context c,float factor));

////get fractal parameter info
//grab center real/imaginary and radius in double precision floating point format
BROT_FUNCTION(bc_get_view_params_as_doubles,void,(bc_context c,double* real,double* img,double* radius));
//ditto, but write text strings. this one exposes the full internal precision.
//buffers sizes for real/img should be at least 64 chars each
BROT_FUNCTION(bc_get_view_params_as_text,void,(bc_context c,char* real,char* img,double* radius));
//get the current maximum iteration depth
BROT_FUNCTION(bc_get_maximum_iterations,uint,(bc_context c));
//store to a 'Place'. That's a little struct, preserves full precision
BROT_FUNCTION(bc_store_location_to,void,(bc_context c,void* place));

////navigation
//zoom on image coords. 0/0 is top left. factors >1 zoom in, factors<1 zoom out
BROT_FUNCTION(bc_zoom_on_window_coord,void,(bc_context c,int x,int y,float factor));
//ditto, but always zoom on image center
BROT_FUNCTION(bc_zoom_centered,void,(bc_context c,float factor));
//move the current view around. deltas are in units of screen width
//eg delta_x=+0.5 moves half an image width to the right
BROT_FUNCTION(bc_move,void,(bc_context c,float delta_x,float delta_y));
//go back to last view
BROT_FUNCTION(bc_history_back,bool,(bc_context c));
//go forward again (prolly after first going back)
BROT_FUNCTION(bc_history_forward,bool,(bc_context c));


///EXPERIMENTAL! DO NOT USE
BROT_FUNCTION(bc_save_all,bool,(bc_context c,const char* filename));
BROT_FUNCTION(bc_save_minimum,bool,(bc_context c,const char* filename));
BROT_FUNCTION(bc_load,bool,(bc_context c,const char* filename));
}

