/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added save_palette - Rolf
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

#include <stdio.h>
#include "common.h"
#include "downfilter.ah"		//needed for RgbImage.h
#include "x86_timer.ah"			//needed for Timing.h
#include "Timing.h"				//needed for RgbImage.h

#include "RgbImage.h"

//internal stuff
#include <math.h>
#include <memory.h>
#include <stdlib.h>

RgbImage::RgbImage(int width,int height):
width(width),height(height),
must_decode(true),
should_decode(true),
aa_degree(-1),
gamma(-1.0f)
{
	data=(unsigned char*)calloc(width*height*3,1);
	set_aa_params(1.5f,1);
}

RgbImage::~RgbImage()
{
	free(data);
}

void
RgbImage::set_palette(const ubyte* palette)
{
	memcpy(this->palette,palette,256*3);
	massage_palette();
}

void
RgbImage::set_size(int width,int height)
{
	if ((this->width!=width)||(this->height!=height))
	{
		SAFE_FREE(data);
		this->width=width;
		this->height=height;
		data=(unsigned char*)calloc(width*height*3,1);
		must_decode=true;
		prep_subsample_info();
	}
}

void
RgbImage::convert_with_palette(const unsigned char* const indexed_image)
{
	//if we don't _have_ to decode, at least deny high fps decoding
	if (!(must_decode||(should_decode&&(timer.elapsed_seconds()>=0.75))))
		return;

	const unsigned char* ip=indexed_image;
	unsigned char* op=data;
	if (aa_degree==1)
	{
		for (int i=0;i<width*height;++i)
		{
			int x=3*(*ip);
			op[0]=palette[x];
			op[1]=palette[x+1];
			op[2]=palette[x+2];
			op+=3;
			++ip;
		}
	}
	else
	if ((aa_degree>1)&&(aa_degree<=8))
	{
		//downfilter linewise
		for (int y=0;y<height;++y)
		{
			x86_downfilter_w_gamma_correction(op,ip,width,&di);
			op+=3*width;
			ip+=aa_degree*aa_degree*width;
		}
	}
	must_decode=false;
	should_decode=false;
	timer.reset();
}

void
RgbImage::set_aa_params(float gamma,int aa_degree)
{
	bool gamma_changed=false;
	bool degree_changed=false;
	//clamp to working (ie supported) range
	if (aa_degree<1) aa_degree=1;
	if (aa_degree>8) aa_degree=8;
	if ((gamma>0.0f)&&(gamma!=this->gamma)) gamma_changed=true;
	if (aa_degree!=this->aa_degree) degree_changed=true;

	if (gamma_changed)
	{
		this->gamma=gamma;
		float exponent=1.0f/gamma;
		uint i;
		for (i=0;i<2048;++i)
		{
			float linear=float(i)/2047u;
			di.gamma_lut[i]=ubyte(255*pow(linear,exponent));
		}
		for (i=0;i<256;++i)
		{
			float linear=float(i)/255u;
			degamma_lut[i]=ushort(32*2047*pow(linear,gamma));
		}
	}
	
	if (degree_changed)
	{
		this->aa_degree=aa_degree;
		prep_subsample_info();
	}
	
	if (degree_changed||gamma_changed) massage_palette();
}

void
RgbImage::massage_palette()
{
	uint divisor=aa_degree*aa_degree;

	for (uint i=0;i<256;++i)
	{
		di.degamma_palette[4*i]=	degamma_lut[palette[3*i]]/divisor;
		di.degamma_palette[4*i+1]=	degamma_lut[palette[3*i+1]]/divisor;
		di.degamma_palette[4*i+2]=	degamma_lut[palette[3*i+2]]/divisor;
		di.degamma_palette[4*i+3]= 0;
	}
	must_decode=true;
}

void
RgbImage::cycle_palette(int cycles)
{
    cycles%=255;
    if (cycles==0) return;
    
    ushort spare_dgp[1024];
	ubyte spare_p[768];
    if (cycles<0)
    {
        memcpy(spare_dgp+4,di.degamma_palette+4*(1-cycles),4*sizeof(ushort)*(255+cycles));
        memcpy(spare_dgp+4*(256+cycles),di.degamma_palette+4,4*sizeof(ushort)*(-cycles));
        memcpy(spare_p+3,palette+3*(1-cycles),3*sizeof(ubyte)*(255+cycles));
        memcpy(spare_p+3*(256+cycles),palette+3,3*sizeof(ubyte)*(-cycles));
    }
    else
    {
        memcpy(spare_dgp+4*(1+cycles),di.degamma_palette+4,4*sizeof(ushort)*(255-cycles));
        memcpy(spare_dgp+4,di.degamma_palette+4*(256-cycles),4*sizeof(ushort)*(cycles));
        memcpy(spare_p+3*(1+cycles),palette+3,3*sizeof(ubyte)*(255-cycles));
        memcpy(spare_p+3,palette+3*(256-cycles),3*sizeof(ubyte)*(cycles));
    }
    memcpy(di.degamma_palette+4,spare_dgp+4,4*255*sizeof(ushort));
    memcpy(palette+3,spare_p+3,3*255*sizeof(ubyte));
    must_decode=true;
}

void
RgbImage::prep_subsample_info()
{
	int x,y;
	int sample_count=0;
	//prepare subsample offset array
	for (y=0;y<aa_degree;++y) for (x=0;x<aa_degree;++x)
	{
		if (x==(aa_degree-1))
		{
			if (y==(aa_degree-1))	di.subpixel_offset[sample_count]=-(aa_degree-1)*aa_degree*width+1;
			else					di.subpixel_offset[sample_count]=aa_degree*(width-1)+1;
		}
		else      					di.subpixel_offset[sample_count]=1;
		++sample_count;
	}
	di.subpixel_count=aa_degree*aa_degree;
}

void
RgbImage::save_palette(FILE* of)
{
	fwrite(palette,sizeof(ubyte),768,of);
}

